from typing import Any, Callable, Iterable, List, Optional, TYPE_CHECKING, Tuple, Union, Dict
import argparse

import matplotlib.pyplot as plt

from sinter import shot_error_rate_to_piece_error_rate
from sinter._main_combine import ExistingData
from sinter._plotting import plot_discard_rate
from sinter._plotting import plot_error_rate

if TYPE_CHECKING:
    import sinter


def parse_args(args: List[str]) -> Any:
    parser = argparse.ArgumentParser(description='Plot collected CSV data.',
                                     prog='sinter plot')
    parser.add_argument('--filter_func',
                        type=str,
                        default="True",
                        help='A python expression that determines whether a case is kept or not.\n'
                             'Available values:\n'
                             '    metadata: The parsed value from the json_metadata column.\n'
                             '    decoder: The decoder that decoded the case.\n'
                             '    strong_id: The cryptographic hash of the case that was sampled.\n'
                             'Expected expression type:\n'
                             '    Something that can be given to `bool` to get True or False.\n'
                             'Examples:\n'
                             '''    --filter_func "decoder=='pymatching'"\n'''
                             '''    --filter_func "0.001 < metadata['p'] < 0.005"\n''')
    parser.add_argument('--x_func',
                        type=str,
                        default="1",
                        help='A python expression that determines where points go on the x axis.\n'
                             'Available values:\n'
                             '    metadata: The parsed value from the json_metadata column.\n'
                             '    decoder: The decoder that decoded the case.\n'
                             '    strong_id: The cryptographic hash of the case that was sampled.\n'
                             'Expected expression type:\n'
                             '    Something that can be given to `float` to get a float.\n'
                             'Examples:\n'
                             '''    --x_func "metadata['p']"\n'''
                             '''    --x_func "metadata['path'].split('/')[-1].split('.')[0]"\n'''
                        )
    parser.add_argument('--fig_size',
                        type=int,
                        nargs=2,
                        default=None,
                        help='Desired figure width in pixels.')
    parser.add_argument('--fig_height',
                        type=int,
                        default=None,
                        help='Desired figure height in pixels.')
    parser.add_argument('--group_func',
                        type=str,
                        default="'all data (use -group_func and -x_func to group into curves)'",
                        help='A python expression that determines how points are grouped into curves.\n'
                             'Available values:\n'
                             '    metadata: The parsed value from the json_metadata column.\n'
                             '    decoder: The decoder that decoded the case.\n'
                             '    strong_id: The cryptographic hash of the case that was sampled.\n'
                             'Expected expression type:\n'
                             '    Something that can be given to `str` to get a useful string.\n'
                             'Examples:\n'
                             '''    --group_func "(decoder, metadata['d'])"\n'''
                             '''    --group_func "metadata['path'].split('/')[-2]"\n'''
                        )
    parser.add_argument('--failure_unit_name',
                        type=str,
                        default=None,
                        help='The unit of failure, typically either "shot" (the default) or "round".\n'
                             'If this argument is specified, --failure_units_per_shot_func must also be specified.\n'''
                        )
    parser.add_argument('--failure_units_per_shot_func',
                        type=str,
                        default=None,
                        help='A python expression that evaluates to the number of failure units there are per shot.\n'
                             'For example, if the failure unit is rounds, this should be an expression that returns\n'
                             'the number of rounds in a shot. Sinter has no way of knowing what you consider a round\n'
                             'to be, otherwise.\n'
                             '\n'
                             'This value is used to rescale the logical error rate plots. For example, if there are 4\n'
                             'failure units per shot then a shot error rate of 10% corresponds to a unit failure rate\n'
                             'of 2.7129%. The conversion formula (assuming <50% error rates) is:\n'
                             '\n'
                             '    P_unit = 0.5 - 0.5 * (1 - 2 * P_shot)**(1/units_per_shot)\n'
                             '\n'
                             'Available values:\n'
                             '    metadata: The parsed value from the json_metadata column.\n'
                             '    decoder: The decoder that decoded the case.\n'
                             '    strong_id: The cryptographic hash of the case that was sampled.\n'
                             '\n'
                             'Expected expression type:\n'
                             '    float.\n'
                             '\n'
                             'Examples:\n'
                             '''    --failure_units_per_shot_func "metadata['rounds']"\n'''
                             '''    --failure_units_per_shot_func "metadata['distance'] * 3"\n'''
                             '''    --failure_units_per_shot_func "10"\n'''
                        )
    parser.add_argument('--plot_args_func',
                        type=str,
                        default='''{'marker': 'ov*sp^<>8PhH+xXDd|'[index % 18]}''',
                        help='A python expression used to customize the look of curves.\n'
                             'Available values:\n'
                             '    index: A unique integer identifying the curve.\n'
                             '    key: The group key (returned from --group_func) identifying the curve.\n'
                             'Expected expression type:\n'
                             '    A dictionary to give to matplotlib plotting functions as a **kwargs argument.\n'
                             'Examples:\n'
                             """    --plot_args_func "'''{'label': 'curve #' + str(index), 'linewidth': 5}'''"\n"""
                             """    --plot_args_func "'''{'marker': 'ov*sp^<>8PhH+xXDd|'[index % 18]}'''"\n"""
                        )
    parser.add_argument('--in',
                        type=str,
                        nargs='+',
                        required=True,
                        help='Input files to get data from.')
    parser.add_argument('--type',
                        choices=['error_rate', 'discard_rate'],
                        nargs='+',
                        default=(),
                        help='Picks the figures to include.')
    parser.add_argument('--out',
                        type=str,
                        default=None,
                        help='Output file to write the plot to.\n'
                             'The file extension determines the type of image.\n'
                             'Either this or --show must be specified.')
    parser.add_argument('--xaxis',
                        type=str,
                        default='[log]',
                        help='Customize the X axis title. Prefix [log] for log scale.')
    parser.add_argument('--show',
                        action='store_true',
                        help='Displays the plot in a window.\n'
                             'Either this or --out must be specified.')
    parser.add_argument('--ymin',
                        default=None,
                        type=float,
                        help='Sets the minimum value of the y axis (max always 1).')
    parser.add_argument('--title',
                        default=None,
                        type=str,
                        help='Sets the title of the plot.')
    parser.add_argument('--highlight_max_likelihood_factor',
                        type=float,
                        default=1000,
                        help='The relative likelihood ratio that determines the color highlights around curves.\n'
                             'Set this to 1 or larger. Set to 1 to disable highlighting.')

    a = parser.parse_args(args=args)
    if not a.show and a.out is None:
        raise ValueError("Must specify '--out file' or '--show'.")
    if (a.failure_units_per_shot_func is not None) != (a.failure_unit_name is not None):
        raise ValueError("--failure_units_per_shot_func and --failure_unit_name can only be specified together.")
    if a.failure_units_per_shot_func is None:
        a.failure_units_per_shot_func = "1"
    if a.failure_unit_name is None:
        a.failure_unit_name = 'shot'

    a.x_func = eval(compile(
        'lambda *, decoder, metadata, strong_id: ' + a.x_func,
        filename='x_func:command_line_arg',
        mode='eval'))
    a.group_func = eval(compile(
        'lambda *, decoder, metadata, strong_id: ' + a.group_func,
        filename='group_func:command_line_arg',
        mode='eval'))
    a.filter_func = eval(compile(
        'lambda *, decoder, metadata, strong_id: ' + a.filter_func,
        filename='filter_func:command_line_arg',
        mode='eval'))
    a.failure_units_per_shot_func = eval(compile(
        'lambda *, decoder, metadata, strong_id: ' + a.failure_units_per_shot_func,
        filename='failure_units_per_shot_func:command_line_arg',
        mode='eval'))
    a.plot_args_func = eval(compile(
        'lambda index, key: ' + a.plot_args_func,
        filename='plot_args_func:command_line_arg',
        mode='eval'))
    return a


def _plot_helper(
    *,
    samples: Union[Iterable['sinter.TaskStats'], ExistingData],
    group_func: Callable[['sinter.TaskStats'], Any],
    filter_func: Callable[['sinter.TaskStats'], Any],
    failure_units_per_shot_func: Callable[['sinter.TaskStats'], Any],
    x_func: Callable[['sinter.TaskStats'], Any],
    failure_unit: str,
    include_discard_rate_plot: bool,
    include_error_rate_plot: bool,
    highlight_max_likelihood_factor: Optional[float],
    xaxis: str,
    min_y: Optional[float],
    title: Optional[str],
    fig_size: Optional[Tuple[int, int]],
    plot_args_func: Callable[[int, Any], Dict[str, Any]],
) -> Tuple[plt.Figure, List[plt.Axes]]:
    if isinstance(samples, ExistingData):
        total = samples
    else:
        total = ExistingData()
        for sample in samples:
            total.add_sample(sample)
    total.data = {k: v
                  for k, v in total.data.items()
                  if bool(filter_func(v))}

    if not include_error_rate_plot and not include_discard_rate_plot:
        include_error_rate_plot = True
        include_discard_rate_plot = any(s.discards for s in total.data.values())

    fig: plt.Figure
    ax_err: Optional[plt.Axes]
    ax_dis: Optional[plt.Axes]
    if include_error_rate_plot and include_discard_rate_plot:
        fig, (ax_err, ax_dis) = plt.subplots(1, 2)
    elif include_error_rate_plot:
        fig, ax_err = plt.subplots(1, 1)
        ax_dis = None
    elif include_discard_rate_plot:
        fig, ax_dis = plt.subplots(1, 1)
        ax_err = None
    else:
        raise NotImplementedError()

    plotted_stats: List['sinter.TaskStats'] = [
        stat
        for stat in total.data.values()
        if filter_func(stat)
    ]

    if xaxis.startswith('[log]'):
        if ax_err is not None:
            ax_err.loglog()
        if ax_dis is not None:
            ax_dis.semilogx()
        xaxis = xaxis[5:]
    else:
        if ax_err is not None:
            ax_err.semilogy()

    if ax_err is not None:
        plot_error_rate(
            ax=ax_err,
            stats=plotted_stats,
            group_func=group_func,
            x_func=x_func,
            failure_units_per_shot_func=failure_units_per_shot_func,
            highlight_max_likelihood_factor=highlight_max_likelihood_factor,
            plot_args_func=plot_args_func,
        )
        if min_y is None:
            min_y = 1
            for stat in plotted_stats:
                if stat.shots <= stat.discards:
                    continue
                err_rate = stat.errors / (stat.shots - stat.discards)
                pieces = failure_units_per_shot_func(stat)
                err_rate = shot_error_rate_to_piece_error_rate(err_rate, pieces=pieces)
                if err_rate < min_y:
                    min_y = err_rate
            if not plotted_stats:
                min_y = 1e-4
            low_d = 4
            while 10**-low_d > min_y*0.9 and low_d < 10:
                low_d += 1
            min_y = 10**-low_d
        major_tick_steps = 1
        while 10**-major_tick_steps >= min_y * 0.1:
            major_tick_steps += 1
        ax_err.set_yticks([10**-d for d in range(major_tick_steps)])
        ax_err.set_yticks([b*10**-d for d in range(1, major_tick_steps) for b in range(2, 10)], minor=True)
        ax_err.set_ylim(min_y, 1e-0)
        ax_err.set_ylabel(f"Logical Error Rate (per {failure_unit})")
        ax_err.grid(which='major', color='#000000')
        ax_err.grid(which='minor', color='#DDDDDD')
        ax_err.legend()
    if ax_dis is not None:
        plot_discard_rate(
            ax=ax_dis,
            stats=plotted_stats,
            group_func=group_func,
            failure_units_per_shot_func=failure_units_per_shot_func,
            x_func=x_func,
            highlight_max_likelihood_factor=highlight_max_likelihood_factor,
            plot_args_func=plot_args_func,
        )
        ax_dis.set_ylim(0, 1)
        ax_err.grid()
        ax_dis.set_ylabel(f"Discard Rate (per {failure_unit}")
        ax_dis.legend()

    if xaxis:
        if ax_err is not None:
            ax_err.set_xlabel(xaxis)
        if ax_dis is not None:
            ax_dis.set_xlabel(xaxis)

    vs_suffix = ''
    if xaxis is not None:
        vs_suffix = f' vs {xaxis}'
    if ax_err is not None:
        ax_err.set_title(f'Logical Error Rate per {failure_unit}{vs_suffix}')
        if title is not None:
            ax_err.set_title(title)
    if ax_dis is not None:
        ax_dis.set_title(f'Discard Rate per {failure_unit}{vs_suffix}')

    if fig_size is None:
        fig.set_size_inches(10 * (include_discard_rate_plot + include_error_rate_plot), 10)
        fig.set_dpi(100)
    else:
        w, h = fig_size
        fig.set_size_inches(w / 100, h / 100)
        fig.set_dpi(100)
    fig.tight_layout()
    axs = [e for e in [ax_err, ax_dis] if e is not None]
    return fig, axs


def main_plot(*, command_line_args: List[str]):
    args = parse_args(command_line_args)
    total = ExistingData()
    for file in getattr(args, 'in'):
        total += ExistingData.from_file(file)

    fig, _ = _plot_helper(
        samples=total,
        group_func=lambda summary: args.group_func(
            decoder=summary.decoder,
            metadata=summary.json_metadata,
            strong_id=summary.strong_id),
        x_func=lambda summary: args.x_func(
            decoder=summary.decoder,
            metadata=summary.json_metadata,
            strong_id=summary.strong_id),
        filter_func=lambda summary: args.filter_func(
            decoder=summary.decoder,
            metadata=summary.json_metadata,
            strong_id=summary.strong_id),
        failure_units_per_shot_func=lambda summary: args.failure_units_per_shot_func(
            decoder=summary.decoder,
            metadata=summary.json_metadata,
            strong_id=summary.strong_id),
        failure_unit=args.failure_unit_name,
        include_discard_rate_plot='discard_rate' in args.type,
        include_error_rate_plot='error_rate' in args.type,
        xaxis=args.xaxis,
        fig_size=args.fig_size,
        min_y=args.ymin,
        highlight_max_likelihood_factor=args.highlight_max_likelihood_factor,
        title=args.title,
        plot_args_func=args.plot_args_func,
    )
    if args.out is not None:
        fig.savefig(args.out)
    if args.show:
        plt.show()



import pathlib

import numpy as np
import stim
import tempfile

import sinter


def test_predict_on_disk_no_postselect():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        dem_path = tmp_dir / 'dem'
        dets_path = tmp_dir / 'dets'
        obs_out_path = tmp_dir / 'obs'

        with open(dem_path, 'w') as f:
            print("""
                error(0.1) D0 L0
                detector(0, 0, 0, 1) D0
            """, file=f)
        with open(dets_path, 'w') as f:
            print("0", file=f)
            print("1", file=f)

        sinter.predict_on_disk(
            decoder='pymatching',
            dem_path=dem_path,
            dets_path=dets_path,
            dets_format='01',
            obs_out_path=obs_out_path,
            obs_out_format='01',
            postselect_detectors_with_non_zero_4th_coord=False,
            discards_out_path=None,
            discards_out_format=None,
        )

        with open(obs_out_path) as f:
            assert f.read() == '0\n1\n'


def test_predict_on_disk_yes_postselect():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        dem_path = tmp_dir / 'dem'
        dets_path = tmp_dir / 'dets'
        obs_out_path = tmp_dir / 'obs'
        discards_out_path = tmp_dir / 'discards'

        with open(dem_path, 'w') as f:
            print("""
                error(0.1) D0 L0
                detector(0, 0, 0, 1) D0
            """, file=f)
        with open(dets_path, 'w') as f:
            print("0", file=f)
            print("1", file=f)

        sinter.predict_on_disk(
            decoder='pymatching',
            dem_path=dem_path,
            dets_path=dets_path,
            dets_format='01',
            obs_out_path=obs_out_path,
            obs_out_format='01',
            postselect_detectors_with_non_zero_4th_coord=True,
            discards_out_path=discards_out_path,
            discards_out_format='01',
        )

        with open(obs_out_path) as f:
            assert f.read() == '0\n'
        with open(discards_out_path) as f:
            assert f.read() == '0\n1\n'


def test_predict_discards_bit_packed_none_postselected():
    dem = stim.DetectorErrorModel("""
        error(0.1) D0 L0
    """)
    actual = sinter.predict_discards_bit_packed(
        dem=dem,
        dets_bit_packed=np.packbits(np.array([
            [False],
            [True],
        ], dtype=np.bool8), bitorder='little', axis=1),
        postselect_detectors_with_non_zero_4th_coord=True,
    )
    np.testing.assert_array_equal(
        actual,
        [False, False],
    )


def test_predict_discards_bit_packed_some_postselected():
    dem = stim.DetectorErrorModel("""
        error(0.1) D0 L0
        detector(0, 0, 0, 1) D0
    """)
    actual = sinter.predict_discards_bit_packed(
        dem=dem,
        dets_bit_packed=np.packbits(np.array([
            [False],
            [True],
        ], dtype=np.bool8), bitorder='little', axis=1),
        postselect_detectors_with_non_zero_4th_coord=True,
    )
    np.testing.assert_array_equal(
        actual,
        [False, True],
    )


def test_predict_observables_bit_packed():
    dem = stim.DetectorErrorModel("""
        error(0.1) D0 L0
    """)

    actual = sinter.predict_observables_bit_packed(
        dem=dem,
        dets_bit_packed=np.packbits(np.array([
            [False],
            [True],
        ], dtype=np.bool8), bitorder='little', axis=1),
        decoder='pymatching',
    )
    np.testing.assert_array_equal(
        np.unpackbits(actual, bitorder='little', count=1, axis=1),
        [
            [False],
            [True],
        ],
    )


def test_predict_observables():
    dem = stim.DetectorErrorModel("""
        error(0.1) D0 L0
    """)

    actual = sinter.predict_observables(
        dem=dem,
        dets=np.packbits(np.array([
            [False],
            [True],
        ], dtype=np.bool8), bitorder='little', axis=1),
        decoder='pymatching',
        bit_pack_result=True,
    )
    np.testing.assert_array_equal(
        np.unpackbits(actual, bitorder='little', count=1, axis=1),
        [
            [False],
            [True],
        ],
    )

    actual = sinter.predict_observables(
        dem=dem,
        dets=np.array([
            [False],
            [True],
        ], dtype=np.bool8),
        decoder='pymatching',
        bit_pack_result=True,
    )
    np.testing.assert_array_equal(
        np.unpackbits(actual, bitorder='little', count=1, axis=1),
        [
            [False],
            [True],
        ],
    )

    actual = sinter.predict_observables(
        dem=dem,
        dets=np.packbits(np.array([
            [False],
            [True],
        ], dtype=np.bool8), bitorder='little', axis=1),
        decoder='pymatching',
        bit_pack_result=False,
    )
    np.testing.assert_array_equal(
        actual,
        [[False], [True]],
    )

    actual = sinter.predict_observables(
        dem=dem,
        dets=np.array([
            [False],
            [True],
        ], dtype=np.bool8),
        decoder='pymatching',
        bit_pack_result=False,
    )
    np.testing.assert_array_equal(
        actual,
        [[False], [True]],
    )

    actual = sinter.predict_observables(
        dem=dem,
        dets=np.packbits(np.array([
            [False],
            [True],
        ], dtype=np.bool8), bitorder='little', axis=1),
        decoder='pymatching',
    )
    np.testing.assert_array_equal(
        actual,
        [[False], [True]],
    )

    actual = sinter.predict_observables(
        dem=dem,
        dets=np.array([
            [False],
            [True],
        ], dtype=np.bool8),
        decoder='pymatching',
    )
    np.testing.assert_array_equal(
        actual,
        [[False], [True]],
    )



import math
from typing import Iterator
from typing import Optional

import stim

from sinter._anon_task_stats import AnonTaskStats
from sinter._existing_data import ExistingData
from sinter._task import Task
from sinter._worker import WorkIn
from sinter._worker import WorkOut


DEFAULT_MAX_BATCH_SECONDS = 120


class CollectionTrackerForSingleTask:
    def __init__(
            self,
            *,
            task: Task,
            circuit_path: str,
            dem_path: str,
            existing_data: ExistingData):
        self.unfilled_task = task
        self.task_strong_id = None
        self.circuit_path = circuit_path
        self.dem_path = dem_path
        self.finished_stats = AnonTaskStats()
        self.existing_data = existing_data
        self.deployed_shots = 0
        self.waiting_for_worker_computing_dem_and_strong_id = False
        self.deployed_processes = 0

        task.circuit.to_file(circuit_path)
        if task.detector_error_model is not None:
            task.detector_error_model.to_file(dem_path)
            self.task_strong_id = task.strong_id()
            existing = self.existing_data.data.get(self.task_strong_id)
            if existing is not None:
                self.finished_stats += existing.to_anon_stats()
        if self.copts.max_shots is None and self.copts.max_errors is None:
            raise ValueError('Neither the task nor the collector specified max_shots or max_errors. Must specify one.')

    @property
    def copts(self):
        return self.unfilled_task.collection_options

    def expected_shots_remaining(
            self, *, safety_factor_on_shots_per_error: float = 1) -> float:
        """Doesn't include deployed shots."""
        result: float = float('inf')

        if self.copts.max_shots is not None:
            result = self.copts.max_shots - self.finished_stats.shots

        if self.finished_stats.errors and self.copts.max_errors is not None:
            shots_per_error = self.finished_stats.shots / self.finished_stats.errors
            errors_left = self.copts.max_errors - self.finished_stats.errors
            result = min(result, errors_left * shots_per_error * safety_factor_on_shots_per_error)

        return result

    def expected_time_per_shot(self) -> Optional[float]:
        if self.finished_stats.shots == 0:
            return None
        return self.finished_stats.seconds / self.finished_stats.shots

    def expected_errors_per_shot(self) -> Optional[float]:
        return (self.finished_stats.errors + 1) / (self.finished_stats.shots + 1)

    def expected_time_remaining(self) -> Optional[float]:
        dt = self.expected_time_per_shot()
        n = self.expected_shots_remaining()
        if dt is None or n == float('inf'):
            return None
        return dt * n

    def work_completed(self, result: WorkOut) -> None:
        if self.waiting_for_worker_computing_dem_and_strong_id:
            self.task_strong_id = result.strong_id
            existing = self.existing_data.data.get(self.task_strong_id)
            if existing is not None:
                self.finished_stats += existing.to_anon_stats()
            self.waiting_for_worker_computing_dem_and_strong_id = False
        else:
            self.deployed_shots -= result.stats.shots
            self.finished_stats += result.stats
        self.deployed_processes -= 1

    def is_done(self) -> bool:
        if self.task_strong_id is None or self.waiting_for_worker_computing_dem_and_strong_id:
            return False
        enough_shots = False
        if self.copts.max_shots is not None and self.finished_stats.shots >= self.copts.max_shots:
            enough_shots = True
        if self.copts.max_errors is not None and self.finished_stats.errors >= self.copts.max_errors:
            enough_shots = True
        return enough_shots and self.deployed_shots == 0

    def iter_batch_size_limits(self, *, desperate: bool) -> Iterator[float]:
        if self.finished_stats.shots == 0:
            if self.deployed_shots > 0:
                yield 0
            elif self.copts.start_batch_size is None:
                yield 100
            else:
                yield self.copts.start_batch_size
            return

        # Do exponential ramp-up of batch sizes.
        yield self.finished_stats.shots * 2

        # Don't go super parallel before reaching other maximums.
        if not desperate:
            yield self.finished_stats.shots * 5 - self.deployed_shots

        # Don't take more shots than requested.
        if self.copts.max_shots is not None:
            yield self.copts.max_shots - self.finished_stats.shots - self.deployed_shots

        # Don't take more errors than requested.
        if self.copts.max_errors is not None:
            errors_left = self.copts.max_errors - self.finished_stats.errors
            errors_left += 2  # oversample once count gets low
            de = self.expected_errors_per_shot()
            yield errors_left / de - self.deployed_shots

        # Don't exceed max batch size.
        if self.copts.max_batch_size is not None:
            yield self.copts.max_batch_size

        # If no maximum on batch size is specified, default to 30s maximum.
        max_batch_seconds = self.copts.max_batch_seconds
        if max_batch_seconds is None and self.copts.max_batch_size is None:
            max_batch_seconds = DEFAULT_MAX_BATCH_SECONDS

        # Try not to exceed max batch duration.
        if max_batch_seconds is not None:
            dt = self.expected_time_per_shot()
            if dt is not None and dt > 0:
                yield max(1, math.floor(max_batch_seconds / dt))

    def next_shot_count(self, *, desperate: bool) -> int:
        return math.ceil(min(self.iter_batch_size_limits(desperate=desperate)))

    def provide_more_work(self, *, desperate: bool) -> Optional[WorkIn]:
        if self.task_strong_id is None:
            if self.waiting_for_worker_computing_dem_and_strong_id:
                return None
            self.waiting_for_worker_computing_dem_and_strong_id = True
            self.deployed_processes += 1
            return WorkIn(
                work_key=None,
                circuit_path=self.circuit_path,
                dem_path=self.dem_path,
                decoder=self.unfilled_task.decoder,
                postselection_mask=self.unfilled_task.postselection_mask,
                postselected_observables_mask=self.unfilled_task.postselected_observables_mask,
                json_metadata=self.unfilled_task.json_metadata,
                strong_id=None,
                num_shots=-1,
            )

        # Wait to have *some* data before starting to sample in parallel.
        num_shots = self.next_shot_count(desperate=desperate)
        if num_shots <= 0:
            return None

        self.deployed_shots += num_shots
        self.deployed_processes += 1
        return WorkIn(
            work_key=None,
            strong_id=self.task_strong_id,
            circuit_path=self.circuit_path,
            dem_path=self.dem_path,
            decoder=self.unfilled_task.decoder,
            postselection_mask=self.unfilled_task.postselection_mask,
            postselected_observables_mask=self.unfilled_task.postselected_observables_mask,
            json_metadata=self.unfilled_task.json_metadata,
            num_shots=num_shots,
        )

    def status(self) -> str:
        t = self.expected_time_remaining()
        if t is not None:
            t /= 60
            t = math.ceil(t)
            t = f'{t}'
        terms = [
            'case: ',
            f'processes={self.deployed_processes}'.ljust(13),
            f'~core_mins_left={t}'.ljust(24),
        ]
        if self.task_strong_id is None:
            terms.append(f'(initializing...) ')
        else:
            if self.copts.max_shots is not None:
                terms.append(f'shots_left={max(0, self.copts.max_shots - self.finished_stats.shots)}'.ljust(20))
            if self.copts.max_errors is not None:
                terms.append(f'errors_left={max(0, self.copts.max_errors - self.finished_stats.errors)}'.ljust(20))
        terms.append(f'{self.unfilled_task.json_metadata}')
        return ''.join(terms)


def next_shot_count(prev_data: AnonTaskStats,
                    cur_batch_size: int,
                    max_batch_size: Optional[int],
                    max_errors: Optional[int],
                    max_shots: Optional[int]) -> int:
    result = cur_batch_size
    if max_shots is not None:
        result = min(result, max_shots - prev_data.shots)
    if prev_data.errors and max_errors is not None:
        result = min(result, math.ceil(1.1 * (max_errors - prev_data.errors) * prev_data.shots / prev_data.errors))
    if max_batch_size is not None:
        result = min(result, max_batch_size)
    return result



import sys
from typing import Optional, List


def main(*, command_line_args: Optional[List[str]] = None):
    if command_line_args is None:
        command_line_args = sys.argv[1:]

    mode = command_line_args[0] if command_line_args else None
    if mode == 'combine':
        from sinter._main_combine import main_combine
        return main_combine(command_line_args=command_line_args[1:])
    if mode == 'collect':
        from sinter._main_collect import main_collect
        return main_collect(command_line_args=command_line_args[1:])
    if mode == 'plot':
        from sinter._main_plot import main_plot
        return main_plot(command_line_args=command_line_args[1:])
    if mode == 'predict':
        from sinter._main_predict import main_predict
        return main_predict(command_line_args=command_line_args[1:])

    want_help = mode in ['help', 'h', '--help', '-help', '-h', '--h']
    if not want_help:
        if command_line_args and not command_line_args[0].startswith('-'):
            print(f"\033[31mUnrecognized command: sinter {command_line_args[0]}\033[0m\n", file=sys.stderr)
        else:
            print(f"\033[31mDidn't specify a command.\033[0m\n", file=sys.stderr)
    print(f"Available commands are:\n"
          f"    sinter collect\n"
          f"    sinter combine\n"
          f"    sinter plot"
          f"", file=sys.stderr)
    if not want_help:
        sys.exit(1)


if __name__ == '__main__':
    main()



import os

import contextlib
import multiprocessing
import pathlib
import tempfile
from typing import cast, Iterable, Optional, Iterator, Tuple, Dict, List

from sinter._collection_options import CollectionOptions
from sinter._existing_data import ExistingData
from sinter._task_stats import TaskStats
from sinter._task import Task
from sinter._anon_task_stats import AnonTaskStats
from sinter._collection_tracker_for_single_task import CollectionTrackerForSingleTask
from sinter._worker import worker_loop, WorkIn, WorkOut


class CollectionWorkManager:
    def __init__(self, *,
                 tasks_iter: Iterator[Task],
                 global_collection_options: CollectionOptions,
                 additional_existing_data: Optional[ExistingData],
                 decoders: Optional[Iterable[str]]):
        self.queue_from_workers: Optional[multiprocessing.Queue] = None
        self.queue_to_workers: Optional[multiprocessing.Queue] = None
        self.additional_existing_data = ExistingData() if additional_existing_data is None else additional_existing_data
        self.tmp_dir: Optional[pathlib.Path] = None
        self.exit_stack: Optional[contextlib.ExitStack] = None

        self.global_collection_options = global_collection_options
        self.decoders: Optional[Tuple[str, ...]] = None if decoders is None else tuple(decoders)
        self.did_work = False

        self.workers: List[multiprocessing.Process] = []
        self.active_collectors: Dict[int, CollectionTrackerForSingleTask] = {}
        self.next_collector_key: int = 0
        self.finished_count = 0
        self.deployed_jobs: Dict[int, WorkIn] = {}
        self.next_job_id = 0

        self.tasks_with_decoder_iter: Iterator[Task] = _iter_tasks_with_assigned_decoders(
            tasks_iter=tasks_iter,
            default_decoders=self.decoders,
            global_collections_options=self.global_collection_options)

    def start_workers(self, num_workers: int) -> None:
        assert self.tmp_dir is not None
        current_method = multiprocessing.get_start_method()
        try:
            # To ensure the child processes do not accidentally share ANY state
            # related to, we use 'spawn' instead of 'fork'.
            multiprocessing.set_start_method('spawn', force=True)
            # Create queues after setting start method to work around a deadlock
            # bug that occurs otherwise.
            self.queue_from_workers = multiprocessing.Queue()
            self.queue_to_workers = multiprocessing.Queue()
            self.queue_from_workers.cancel_join_thread()
            self.queue_to_workers.cancel_join_thread()

            num_cpus = os.cpu_count()
            for index in range(num_workers):
                w = multiprocessing.Process(
                    target=worker_loop,
                    args=(self.tmp_dir, self.queue_to_workers, self.queue_from_workers, index % num_cpus))
                w.start()
                self.workers.append(w)
        finally:
            multiprocessing.set_start_method(current_method, force=True)

    def __enter__(self):
        self.exit_stack = contextlib.ExitStack().__enter__()
        self.tmp_dir = pathlib.Path(self.exit_stack.enter_context(tempfile.TemporaryDirectory()))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shut_down_workers()
        self.exit_stack.__exit__(exc_type, exc_val, exc_tb)
        self.exit_stack = None
        self.tmp_dir = None

    def shut_down_workers(self) -> None:
        removed_workers = self.workers
        self.workers = []

        # SIGKILL everything.
        for w in removed_workers:
            # This is supposed to be safe because all state on disk was put
            # in the specified tmp directory which we will handle deleting.
            w.kill()

    def fill_work_queue(self) -> bool:
        while len(self.deployed_jobs) < len(self.workers):
            work = self.provide_more_work()
            if work is None:
                break
            self.did_work = True
            self.queue_to_workers.put(work.with_work_key((self.next_job_id, work.work_key)))
            self.deployed_jobs[self.next_job_id] = work
            self.next_job_id += 1
        return bool(self.deployed_jobs)

    def wait_for_next_sample(self,
                             *,
                             timeout: Optional[float] = None,
                             ) -> TaskStats:
        result = self.queue_from_workers.get(timeout=timeout)
        assert isinstance(result, WorkOut)
        if result.msg_error is not None:
            msg, error = result.msg_error
            if isinstance(error, KeyboardInterrupt):
                raise KeyboardInterrupt()
            raise RuntimeError(f"Worker failed: {msg}") from error

        else:
            job_id, sub_key = result.work_key
            stats = result.stats
            work_in = self.deployed_jobs[job_id]

            self.work_completed(WorkOut(
                work_key=sub_key,
                stats=stats,
                strong_id=result.strong_id,
                msg_error=result.msg_error,
            ))
            del self.deployed_jobs[job_id]
            if stats is None:
                stats = AnonTaskStats()
            return TaskStats(
                strong_id=result.strong_id,
                decoder=work_in.decoder,
                json_metadata=work_in.json_metadata,
                shots=stats.shots,
                errors=stats.errors,
                discards=stats.discards,
                seconds=stats.seconds,
            )

    def _iter_draw_collectors(self, *, prefer_started: bool) -> Iterator[Tuple[int, CollectionTrackerForSingleTask]]:
        if prefer_started:
            yield from self.active_collectors.items()
        while True:
            key = self.next_collector_key
            try:
                task = next(self.tasks_with_decoder_iter)
            except StopIteration:
                break
            collector = CollectionTrackerForSingleTask(
                task=task,
                circuit_path=str((self.tmp_dir / f'circuit_{self.next_collector_key}.stim').absolute()),
                dem_path=str((self.tmp_dir / f'dem_{self.next_collector_key}.dem').absolute()),
                existing_data=self.additional_existing_data,
            )
            if collector.is_done():
                self.finished_count += 1
                continue
            self.next_collector_key += 1
            self.active_collectors[key] = collector
            yield key, collector
        if not prefer_started:
            yield from self.active_collectors.items()

    def is_done(self) -> bool:
        return len(self.active_collectors) == 0

    def work_completed(self, result: WorkOut):
        assert isinstance(result.work_key, int)
        collector_index = cast(int, result.work_key)
        collector = self.active_collectors[collector_index]
        collector.work_completed(result)
        if collector.is_done():
            self.finished_count += 1
            del self.active_collectors[collector_index]

    def provide_more_work(self) -> Optional[WorkIn]:
        iter_collectors = self._iter_draw_collectors(
                prefer_started=len(self.active_collectors) >= 2)
        for desperate in False, True:
            for collector_index, collector in iter_collectors:
                w = collector.provide_more_work(desperate=desperate)
                if w is not None:
                    assert w.work_key is None
                    return w.with_work_key(collector_index)
        return None

    def status(self, *, num_circuits: Optional[int]) -> str:
        if self.is_done():
            if self.did_work:
                main_status = 'Done collecting'
            else:
                main_status = 'There was nothing additional to collect'
        elif num_circuits is not None:
            main_status = f'{num_circuits - self.finished_count} cases left'
        else:
            main_status = "Running..."
        collector_statuses = [
            '\n    ' + collector.status()
            for collector in self.active_collectors.values()
        ]
        if len(collector_statuses) > 24:
            collector_statuses = collector_statuses[:24] + ['\n...']
        return main_status + ''.join(collector_statuses)


def _iter_tasks_with_assigned_decoders(
    *,
    tasks_iter: Iterator[Task],
    default_decoders: Optional[Iterable[str]],
    global_collections_options: CollectionOptions,
) -> Iterator[Task]:
    for task in tasks_iter:
        if task.decoder is None and default_decoders is None:
            raise ValueError("Decoders to use was not specified. decoders is None and task.decoder is None")
        task_decoders = []
        if default_decoders is not None:
            task_decoders.extend(default_decoders)
        if task.decoder is not None and task.decoder not in task_decoders:
            task_decoders.append(task.decoder)
        for decoder in task_decoders:
            yield Task(
                circuit=task.circuit,
                decoder=decoder,
                detector_error_model=task.detector_error_model,
                postselection_mask=task.postselection_mask,
                postselected_observables_mask=task.postselected_observables_mask,
                json_metadata=task.json_metadata,
                collection_options=task.collection_options.combine(global_collections_options),
                skip_validation=True,
            )



import sys
from typing import List

from sinter._csv_out import CSV_HEADER
from sinter._existing_data import ExistingData


def main_combine(*, command_line_args: List[str]):
    if command_line_args:
        total = ExistingData()
        for path in command_line_args:
            total += ExistingData.from_file(path)
    else:
        total = ExistingData.from_file(sys.stdin)

    print(CSV_HEADER)
    for value in total.data.values():
        print(value.to_csv_line())



import collections

import pytest
import stim

import sinter


def test_iter_collect():
    result = collections.defaultdict(sinter.AnonTaskStats)
    for sample in sinter.iter_collect(
        num_workers=2,
        tasks=[
            sinter.Task(
                circuit=stim.Circuit.generated(
                    'repetition_code:memory',
                    rounds=3,
                    distance=3,
                    after_clifford_depolarization=p),
                decoder='pymatching',
                json_metadata={'p': p},
                collection_options=sinter.CollectionOptions(
                    max_shots=1000,
                    max_errors=100,
                    start_batch_size=100,
                    max_batch_size=1000,
                ),
            )
            for p in [0.01, 0.02, 0.03, 0.04]
        ],
    ):
        for stats in sample.new_stats:
            result[stats.json_metadata['p']] += stats.to_anon_stats()
    assert len(result) == 4
    for k, v in result.items():
        assert v.shots >= 1000 or v.errors >= 100
        assert v.discards == 0
    assert result[0.01].errors <= 10
    assert result[0.02].errors <= 30
    assert result[0.03].errors <= 70
    assert 1 <= result[0.04].errors <= 100


def test_collect():
    results = sinter.collect(
        num_workers=2,
        tasks=[
            sinter.Task(
                circuit=stim.Circuit.generated(
                    'repetition_code:memory',
                    rounds=3,
                    distance=3,
                    after_clifford_depolarization=p),
                decoder='pymatching',
                json_metadata={'p': p},
                collection_options=sinter.CollectionOptions(
                    max_shots=1000,
                    max_errors=100,
                    start_batch_size=100,
                    max_batch_size=1000,
                ),
            )
            for p in [0.01, 0.02, 0.03, 0.04]
        ]
    )
    probabilities = [e.json_metadata['p'] for e in results]
    assert len(probabilities) == len(set(probabilities))
    d = {e.json_metadata['p']: e for e in results}
    assert len(d) == 4
    for k, v in d.items():
        assert v.shots >= 1000 or v.errors >= 100
        assert v.discards == 0
    assert d[0.01].errors <= 10
    assert d[0.02].errors <= 30
    assert d[0.03].errors <= 70
    assert 1 <= d[0.04].errors <= 100


def test_iter_collect_list():
    result = collections.defaultdict(sinter.AnonTaskStats)
    for sample in sinter.iter_collect(
        num_workers=2,
        tasks=[
            sinter.Task(
                circuit=stim.Circuit.generated(
                    'repetition_code:memory',
                    rounds=3,
                    distance=3,
                    after_clifford_depolarization=p),
                decoder='pymatching',
                json_metadata={'p': p},
                collection_options=sinter.CollectionOptions(
                    max_errors=100,
                    max_shots=1000,
                    start_batch_size=100,
                    max_batch_size=1000,
                ),
            )
            for p in [0.01, 0.02, 0.03, 0.04]
        ],
    ):
        for stats in sample.new_stats:
            result[stats.json_metadata['p']] += stats.to_anon_stats()
    assert len(result) == 4
    for k, v in result.items():
        assert v.shots >= 1000 or v.errors >= 100
        assert v.discards == 0
    assert result[0.01].errors <= 10
    assert result[0.02].errors <= 30
    assert result[0.03].errors <= 70
    assert 1 <= result[0.04].errors <= 100


def test_iter_collect_worker_fails():
    with pytest.raises(RuntimeError, match="Worker failed"):
        _ = list(sinter.iter_collect(
            decoders=['NOT A VALID DECODER'],
            num_workers=1,
            tasks=iter([
                sinter.Task(
                    circuit=stim.Circuit.generated('repetition_code:memory', rounds=3, distance=3),
                    collection_options=sinter.CollectionOptions(
                        max_errors=1,
                        max_shots=1,
                    ),
                ),
            ]),
        ))



import contextlib
import io
import pathlib
import tempfile

import stim

import pytest
import sinter
from sinter._main import main
from sinter._main_combine import ExistingData
from sinter._plotting import split_by


def test_split_by():
    assert split_by('abcdefcccghi', lambda e: e == 'c') == [list('ab'), list('c'), list('def'), list('ccc'), list('ghi')]


def test_main_collect():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        for distance in [3, 5, 7]:
            c = stim.Circuit.generated(
                'repetition_code:memory',
                rounds=3,
                distance=distance,
                after_clifford_depolarization=0.02)
            with open(d / f'{distance}.stim', 'w') as f:
                print(c, file=f)

        # Collects requested stats.
        main(command_line_args=[
            "collect",
            "--circuits",
            str(d / "3.stim"),
            str(d / "5.stim"),
            str(d / "7.stim"),
            "--max_shots",
            "1000",
            "--max_errors",
            "10",
            "--decoders",
            "pymatching",
            "--processes",
            "4",
            "--quiet",
            "--save_resume_filepath",
            str(d / "out.csv"),
        ])
        data = ExistingData.from_file(d / "out.csv").data
        assert len(data) == 3
        for k, v in data.items():
            assert v.discards == 0
            assert v.errors <= 50
            assert v.shots >= 1000

        # No more work when existing stats at merge location are sufficient.
        with open(d / "out.csv") as f:
            contents1 = f.read()
        main(command_line_args=[
            "collect",
            "--circuits",
            str(d / "3.stim"),
            str(d / "5.stim"),
            str(d / "7.stim"),
            "--max_shots",
            "1000",
            "--max_errors",
            "10",
            "--decoders",
            "pymatching",
            "--processes",
            "4",
            "--quiet",
            "--save_resume_filepath",
            str(d / "out.csv"),
        ])
        with open(d / "out.csv") as f:
            contents2 = f.read()
        assert contents1 == contents2

        # No more work when existing work is sufficient.
        main(command_line_args=[
            "collect",
            "--circuits",
            str(d / "3.stim"),
            str(d / "5.stim"),
            str(d / "7.stim"),
            "--max_shots",
            "1000",
            "--max_errors",
            "10",
            "--decoders",
            "pymatching",
            "--processes",
            "4",
            "--quiet",
            "--existing_data_filepaths",
            str(d / "out.csv"),
            "--save_resume_filepath",
            str(d / "out2.csv"),
        ])
        data2 = ExistingData.from_file(d / "out2.csv").data
        assert len(data2) == 0


def test_main_collect_post_select_observables():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'circuit.stim', 'w') as f:
            print("""
                M(0.125) 0 1
                OBSERVABLE_INCLUDE(0) rec[-1]
                OBSERVABLE_INCLUDE(11) rec[-1] rec[-2]
            """, file=f)

        # Collects requested stats.
        main(command_line_args=[
            "collect",
            "--postselected_observables_predicate",
            "index == 11",
            "--circuits",
            str(d / "circuit.stim"),
            "--max_shots",
            "10000",
            "--max_errors",
            "10000",
            "--decoders",
            "pymatching",
            "--processes",
            "4",
            "--quiet",
            "--save_resume_filepath",
            str(d / "out.csv"),
        ])
        data = sinter.stats_from_csv_files(d / "out.csv")
        assert len(data) == 1
        stats, = data
        assert stats.shots == 10000
        assert 0.21875 - 0.1 < stats.discards / stats.shots < 0.21875 + 0.1
        assert 0.015625 - 0.01 <= stats.errors / (stats.shots - stats.discards) <= 0.015625 + 0.02


def test_main_collect_comma_separated_key_values():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        paths = []
        for distance in [3, 5, 7]:
            c = stim.Circuit.generated(
                'repetition_code:memory',
                rounds=3,
                distance=distance,
                after_clifford_depolarization=0.02)
            path = d / f'd={distance},p=0.02,r=3.0,c=rep_code.stim'
            paths.append(str(path))
            with open(path, 'w') as f:
                print(c, file=f)

        # Collects requested stats.
        main(command_line_args=[
            "collect",
            "--circuits",
            *paths,
            "--max_shots",
            "1000",
            "--metadata_func",
            "sinter.comma_separated_key_values(path)",
            "--max_errors",
            "10",
            "--decoders",
            "pymatching",
            "--processes",
            "4",
            "--quiet",
            "--save_resume_filepath",
            str(d / "out.csv"),
        ])
        data = sinter.stats_from_csv_files(d / "out.csv")
        seen_metadata = frozenset(repr(e.json_metadata) for e in data)
        assert seen_metadata == frozenset([
            "{'c': 'rep_code', 'd': 3, 'p': 0.02, 'r': 3.0}",
            "{'c': 'rep_code', 'd': 5, 'p': 0.02, 'r': 3.0}",
            "{'c': 'rep_code', 'd': 7, 'p': 0.02, 'r': 3.0}",
        ])


def test_main_combine():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.csv', 'w') as f:
            print("""
shots,errors,discards,seconds,decoder,strong_id,json_metadata
300,1,20,1.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
300,100,200,2.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
9,5,4,6.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
            """.strip(), file=f)

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "combine",
                str(d / "input.csv"),
            ])
        assert out.getvalue() == """     shots,    errors,  discards, seconds,decoder,strong_id,json_metadata
       600,       101,       220,    3.00,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
         9,         5,         4,    6.00,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
"""

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "combine",
                str(d / "input.csv"),
                str(d / "input.csv"),
            ])
        assert out.getvalue() == """     shots,    errors,  discards, seconds,decoder,strong_id,json_metadata
      1200,       202,       440,    6.00,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
        18,        10,         8,    12.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
"""


def test_main_plot():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.csv', 'w') as f:
            print("""
shots,errors,discards,seconds,decoder,strong_id,json_metadata
300,1,20,1.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
300,100,200,2.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
9,5,4,6.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
            """.strip(), file=f)

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "plot",
                "--in",
                str(d / "input.csv"),
                "--out",
                str(d / "output.png"),
                "--x_func",
                "int('a' in metadata['path'])",
                "--group_func",
                "decoder",
                "--ymin",
                "1e-3",
                "--title",
                "test_plot",
            ])
        assert (d / "output.png").exists()


def test_main_plot_2():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.csv', 'w') as f:
            print("""
shots,errors,discards,seconds,decoder,strong_id,json_metadata
300,1,20,1.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
300,100,200,2.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
9,5,4,6.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
            """.strip(), file=f)

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "plot",
                "--in",
                str(d / "input.csv"),
                "--out",
                str(d / "output.png"),
                "--x_func",
                "int('a' in metadata['path'])",
                "--group_func",
                "decoder",
                "--plot_args_func",
                "{'lw': 2}",
            ])
        assert (d / "output.png").exists()


def test_main_plot_failure_units():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.csv', 'w') as f:
            print("""
shots,errors,discards,seconds,decoder,strong_id,json_metadata
300,1,20,1.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""r"":15,""d"":5}"
300,100,200,2.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf7,"{""r"":9,""d"":3}"
9,5,4,6.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""r"":6,""d"":2}"
            """.strip(), file=f)

        out = io.StringIO()
        with pytest.raises(ValueError,  match="specified together"):
            main(command_line_args=[
                "plot",
                "--in",
                str(d / "input.csv"),
                "--out",
                str(d / "output.png"),
                "--x_func",
                "metadata['d']",
                "--group_func",
                "decoder",
                "--failure_units_per_shot_func",
                "metadata['r']",
            ])
        with pytest.raises(ValueError,  match="specified together"):
            main(command_line_args=[
                "plot",
                "--in",
                str(d / "input.csv"),
                "--out",
                str(d / "output.png"),
                "--x_func",
                "metadata['d']",
                "--group_func",
                "decoder",
                "--failure_unit_name",
                "Rounds",
            ])
        assert not (d / "output.png").exists()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "plot",
                "--in",
                str(d / "input.csv"),
                "--out",
                str(d / "output.png"),
                "--x_func",
                "metadata['d']",
                "--group_func",
                "decoder",
                "--failure_units_per_shot_func",
                "metadata['r']",
                "--failure_unit_name",
                "Rounds",
            ])
        assert (d / "output.png").exists()


def test_main_predict():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.dets', 'w') as f:
            print("""
shot D0
shot
            """, file=f)
        with open(d / f'input.dem', 'w') as f:
            print("""
error(0.1) D0 L0
            """, file=f)

        main(command_line_args=[
            "predict",
            "--dets",
            str(d / "input.dets"),
            "--dem",
            str(d / "input.dem"),
            "--decoder",
            "pymatching",
            "--dets_format",
            "dets",
            "--obs_out",
            str(d / "output.01"),
            "--obs_out_format",
            "01",
        ])
        with open(d / 'output.01') as f:
            assert f.read() == '1\n0\n'



import numpy as np
import pytest
import stim

from sinter._collection import post_selection_mask_from_4th_coord
from sinter._decoding import sample_decode

DECODER_PACKAGES = [
    ('pymatching', 'pymatching'),
    ('fusion_blossom', 'fusion_blossom'),
    ('internal', 'gqec'),
    ('internal_correlated', 'gqec'),
]


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_decode_repetition_code(decoder: str, required_import: str):
    pytest.importorskip(required_import)

    circuit = stim.Circuit.generated('repetition_code:memory',
                                     rounds=3,
                                     distance=3,
                                     after_clifford_depolarization=0.05)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert 1 <= result.errors <= 100
    assert result.shots == 1000


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_decode_surface_code(decoder: str, required_import: str):
    pytest.importorskip(required_import)

    circuit = stim.Circuit.generated(
        "surface_code:rotated_memory_x",
        distance=3,
        rounds=15,
        after_clifford_depolarization=0.001,
    )
    stats = sample_decode(
        num_shots=1000,
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        decoder=decoder,
    )
    assert 0 <= stats.errors <= 50


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_empty(decoder: str, required_import: str):
    pytest.importorskip(required_import)
    circuit = stim.Circuit()
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert result.shots == 1000
    assert result.errors == 0


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_no_observables(decoder: str, required_import: str):
    pytest.importorskip(required_import)
    circuit = stim.Circuit("""
        X_ERROR(0.1) 0
        M 0
        DETECTOR rec[-1]
    """)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert result.shots == 1000
    assert result.errors == 0


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_invincible_observables(decoder: str, required_import: str):
    pytest.importorskip(required_import)
    circuit = stim.Circuit("""
        X_ERROR(0.1) 0
        M 0 1
        DETECTOR rec[-2]
        OBSERVABLE_INCLUDE(1) rec[-1]
    """)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert result.shots == 1000
    assert result.errors == 0


@pytest.mark.parametrize('decoder,required_import,offset', [(a, b, c) for a, b in DECODER_PACKAGES for c in range(8)])
def test_observable_offsets_mod8(decoder: str, required_import: str, offset: int):
    pytest.importorskip(required_import)
    circuit = stim.Circuit("""
        X_ERROR(0.1) 0
        MR 0
        DETECTOR rec[-1]
    """) * (8 + offset) + stim.Circuit("""
        X_ERROR(0.1) 0
        MR 0
        OBSERVABLE_INCLUDE(0) rec[-1]
    """)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert result.shots == 1000
    assert 50 <= result.errors <= 150


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_no_detectors(decoder: str, required_import: str):
    pytest.importorskip(required_import)
    circuit = stim.Circuit("""
        X_ERROR(0.1) 0
        M 0
        OBSERVABLE_INCLUDE(0) rec[-1]
    """)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert 50 <= result.errors <= 150


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_no_detectors_with_post_mask(decoder: str, required_import: str):
    pytest.importorskip(required_import)
    circuit = stim.Circuit("""
        X_ERROR(0.1) 0
        M 0
        OBSERVABLE_INCLUDE(0) rec[-1]
    """)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        post_mask=np.array([], dtype=np.uint8),
        num_shots=1000,
        decoder=decoder,
    )
    assert result.discards == 0
    assert 50 <= result.errors <= 150


@pytest.mark.parametrize('decoder,required_import', DECODER_PACKAGES)
def test_post_selection(decoder: str, required_import: str):
    pytest.importorskip(required_import)
    circuit = stim.Circuit("""
        X_ERROR(0.6) 0
        M 0
        DETECTOR(2, 0, 0, 1) rec[-1]
        OBSERVABLE_INCLUDE(0) rec[-1]

        X_ERROR(0.5) 1
        M 1
        DETECTOR(1, 0, 0) rec[-1]
        OBSERVABLE_INCLUDE(0) rec[-1]
        
        X_ERROR(0.1) 2
        M 2
        OBSERVABLE_INCLUDE(0) rec[-1]
    """)
    result = sample_decode(
        circuit_obj=circuit,
        circuit_path=None,
        dem_obj=circuit.detector_error_model(decompose_errors=True),
        dem_path=None,
        post_mask=post_selection_mask_from_4th_coord(circuit),
        num_shots=2000,
        decoder=decoder,
    )
    assert 1050 <= result.discards <= 1350
    assert 40 <= result.errors <= 160



import io

from matplotlib import pyplot as plt

import sinter


def test_better_sorted_str_terms():
    f = sinter.better_sorted_str_terms
    assert f('everyone et al. 2020') == ('everyone et al', '.', ' ', 2020)
    assert f('a') == ('a',)
    assert f('abc') == ('abc',)
    assert f('a1b2') == ('a', 1, 'b', 2)
    assert f('a1.5b2') == ('a', 1.5, 'b', 2)
    assert f('a1.5.3b2') == ('a', (1, 5, 3), 'b', 2)
    assert sorted([
        "planar d=10 r=30",
        "planar d=16 r=36",
        "planar d=4 r=12",
        "toric d=10 r=30",
        "toric d=18 r=54",
    ], key=f) == [
        "planar d=4 r=12",
        "planar d=10 r=30",
        "planar d=16 r=36",
        "toric d=10 r=30",
        "toric d=18 r=54",
    ]


def test_plotting_does_not_crash():
    data = io.StringIO()
    data.write("""  shots,errors,discards,seconds,decoder,strong_id,json_metadata
1000000, 837,0,36.6,pymatching,9f7e20c54fec45b6aef7491b774dd5c0a3b9a005aa82faf5b9c051d6e40d60a9,"{""d"":3,""p"":0.001}"
  53498,1099,0,6.52,pymatching,3f40432443a99b933fb548b831fb54e7e245d9d73a35c03ea5a2fb2ce270f8c8,"{""d"":3,""p"":0.005}"
  16269,1023,0,3.23,pymatching,17b2e0c99560d20307204494ac50e31b33e50721b4ebae99d9e3577ae7248874,"{""d"":3,""p"":0.01}"
1000000, 151,0,77.3,pymatching,e179a18739201250371ffaae0197d8fa19d26b58dfc2942f9f1c85568645387a,"{""d"":5,""p"":0.001}"
  11363,1068,0,12.5,pymatching,a4dec28934a033215ff1389651a26114ecc22016a6e122008830cf7dd04ba5ad,"{""d"":5,""p"":0.01}"
  61569,1001,0,24.5,pymatching,2fefcc356752482fb4c6d912c228f6d18762f5752796c668b6abeb7775f5de92,"{""d"":5,""p"":0.005}"
""")
    data.seek(0)
    stats = sinter.stats_from_csv_files(data)

    fig, ax = plt.subplots(1, 1)
    sinter.plot_error_rate(
        ax=ax,
        stats=stats,
        group_func=lambda e: f"Rotated Surface Code d={e.json_metadata['d']}",
        x_func=lambda e: e.json_metadata['p'],
        plot_args_func=lambda k, e: {'marker': "ov*sp^<>8PhH+xXDd|"[k]},
    )
    sinter.plot_error_rate(
        ax=ax,
        stats=stats,
        group_func=lambda e: f"Rotated Surface Code d={e.json_metadata['d']}",
        x_func=lambda e: e.json_metadata['p'],
        plot_args_func=lambda k, e: {'marker': "ov*sp^<>8PhH+xXDd|"[k]},
        failure_units_per_shot_func=lambda stats: stats.json_metadata['d'] * 3,
    )
    sinter.plot_discard_rate(
        ax=ax,
        stats=stats,
        group_func=lambda e: f"Rotated Surface Code d={e.json_metadata['d']}",
        x_func=lambda e: e.json_metadata['p'],
        plot_args_func=lambda k, e: {'marker': "ov*sp^<>8PhH+xXDd|"[k]},
    )
    sinter.plot_discard_rate(
        ax=ax,
        stats=stats,
        group_func=lambda e: f"Rotated Surface Code d={e.json_metadata['d']}",
        x_func=lambda e: e.json_metadata['p'],
        plot_args_func=lambda k, e: {'marker': "ov*sp^<>8PhH+xXDd|"[k]},
        failure_units_per_shot_func=lambda stats: stats.json_metadata['d'] * 3,
    )


def test_group_by():
    assert sinter.group_by([1, 2, 3], key=lambda i: i == 2) == {False: [1, 3], True: [2]}
    assert sinter.group_by(range(10), key=lambda i: i % 3) == {0: [0, 3, 6, 9], 1: [1, 4, 7], 2: [2, 5, 8]}
    assert sinter.group_by([], key=lambda i: 0) == {}
    assert sinter.group_by([1, 2, 3, 1], key=lambda i: 0) == {0: [1, 2, 3, 1]}



import os

from typing import Any, Optional, Tuple, TYPE_CHECKING
import tempfile

if TYPE_CHECKING:
    import multiprocessing
    import numpy as np
    import pathlib
    import sinter
    import stim
    from sinter._json_type import JSON_TYPE


class WorkIn:
    def __init__(
            self,
            *,
            work_key: Any,
            circuit_path: str,
            dem_path: str,
            decoder: str,
            strong_id: Optional[str],
            postselection_mask: 'Optional[np.ndarray]',
            postselected_observables_mask: 'Optional[np.ndarray]',
            json_metadata: 'JSON_TYPE',
            num_shots: int):
        self.work_key = work_key
        self.circuit_path = circuit_path
        self.dem_path = dem_path
        self.decoder = decoder
        self.strong_id = strong_id
        self.postselection_mask = postselection_mask
        self.postselected_observables_mask = postselected_observables_mask
        self.json_metadata = json_metadata
        self.num_shots = num_shots

    def with_work_key(self, work_key: Any) -> 'WorkIn':
        return WorkIn(
            work_key=work_key,
            circuit_path=self.circuit_path,
            dem_path=self.dem_path,
            decoder=self.decoder,
            postselection_mask=self.postselection_mask,
            postselected_observables_mask=self.postselected_observables_mask,
            json_metadata=self.json_metadata,
            strong_id=self.strong_id,
            num_shots=self.num_shots,
        )


def auto_dem(circuit: 'stim.Circuit') -> 'stim.DetectorErrorModel':
    return circuit.detector_error_model(
        allow_gauge_detectors=False,
        approximate_disjoint_errors=True,
        block_decomposition_from_introducing_remnant_edges=False,
        decompose_errors=True,
        flatten_loops=True,
        ignore_decomposition_failures=False,
    )


class WorkOut:
    def __init__(
            self,
            *,
            work_key: Any,
            stats: Optional['sinter.AnonTaskStats'],
            strong_id: str,
            msg_error: Optional[Tuple[str, BaseException]]):
        self.work_key = work_key
        self.stats = stats
        self.strong_id = strong_id
        self.msg_error = msg_error


def worker_loop(tmp_dir: 'pathlib.Path',
                inp: 'multiprocessing.Queue',
                out: 'multiprocessing.Queue',
                core_affinity: Optional[int]) -> None:
    try:
        if core_affinity is not None and hasattr(os, 'sched_setaffinity'):
            os.sched_setaffinity(0, {core_affinity})
    except:
        # If setting the core affinity fails, we keep going regardless.
        pass

    try:
        with tempfile.TemporaryDirectory(dir=tmp_dir) as child_dir:
            while True:
                work: Optional[WorkIn] = inp.get()
                if work is None:
                    return
                out.put(do_work_safely(work, child_dir))
    except KeyboardInterrupt:
        pass


def do_work_safely(work: WorkIn, child_dir: str) -> WorkOut:
    try:
        return do_work(work, child_dir)
    except BaseException as ex:
        import traceback
        return WorkOut(
            work_key=work.work_key,
            stats=None,
            strong_id=work.strong_id,
            msg_error=(traceback.format_exc(), ex),
        )


def do_work(work: WorkIn, child_dir: str) -> WorkOut:
    import stim
    from sinter._task import Task
    from sinter._decoding import sample_decode

    if work.strong_id is None:
        circuit = stim.Circuit.from_file(work.circuit_path)
        dem = auto_dem(circuit)
        dem.to_file(work.dem_path)

        task = Task(
            circuit=circuit,
            decoder=work.decoder,
            detector_error_model=dem,
            postselection_mask=work.postselection_mask,
            postselected_observables_mask=work.postselected_observables_mask,
            json_metadata=work.json_metadata,
        )

        return WorkOut(
            work_key=work.work_key,
            stats=None,
            strong_id=task.strong_id(),
            msg_error=None,
        )

    stats: 'sinter.AnonTaskStats' = sample_decode(
        num_shots=work.num_shots,
        circuit_path=work.circuit_path,
        circuit_obj=None,
        dem_path=work.dem_path,
        dem_obj=None,
        post_mask=work.postselection_mask,
        postselected_observable_mask=work.postselected_observables_mask,
        decoder=work.decoder,
        tmp_dir=child_dir,
    )

    return WorkOut(
        stats=stats,
        work_key=work.work_key,
        strong_id=work.strong_id,
        msg_error=None,
    )



import pathlib

import stim


def decode_using_internal_decoder(*,
                                  num_shots: int,
                                  num_dets: int,
                                  num_obs: int,
                                  dem_path: pathlib.Path,
                                  dets_b8_in_path: pathlib.Path,
                                  obs_predictions_b8_out_path: pathlib.Path,
                                  tmp_dir: pathlib.Path,
                                  use_correlated_decoding: bool,
                                  ) -> None:
    """Use internal decoder to predict observables from detection events."""

    if num_dets == 0:
        with open(obs_predictions_b8_out_path, 'wb') as f:
            f.write(b'\0' * (num_obs * num_shots))
        return

    try:
        import gqec  # Internal python wheel.
    except ImportError as ex:
        raise ImportError(
            "The decoder 'internal' isn't installed\n"
            "The decoder 'internal_correlated' isn't installed\n"
            "These decoders aren't publicly available.\n"
        ) from ex

    gqec.run_finite_match_main(
        dem_filepath=str(dem_path),
        output_type="predictions:b8",
        output_filepath=str(obs_predictions_b8_out_path),
        dets_has_observables=False,
        dets_format='b8',
        dets_filepath=str(dets_b8_in_path),
        ignore_distance_1_errors=True,
        ignore_undecomposed_errors=True,
        use_correlated_decoding=use_correlated_decoding,
    )



import pathlib

from sinter._decoding_pymatching_v1 import decode_using_pymatching_v1
from sinter._decoding_pymatching_v2 import decode_using_pymatching_v2


_method_to_use = None


def decode_using_pymatching(*,
                            num_shots: int,
                            num_dets: int,
                            num_obs: int,
                            dem_path: pathlib.Path,
                            dets_b8_in_path: pathlib.Path,
                            obs_predictions_b8_out_path: pathlib.Path,
                            tmp_dir: pathlib.Path,
                            ) -> None:

    try:
        import pymatching
    except ImportError as ex:
        raise ImportError(
            "The decoder 'pymatching' isn't installed\n"
            "To fix this, install the python package 'pymatching' into your environment.\n"
            "For example, if you are using pip, run `pip install pymatching`.\n"
        ) from ex

    global _method_to_use
    if _method_to_use is None:
        if getattr(getattr(pymatching, '_cpp_pymatching', None), 'main', None) is not None:
            _method_to_use = decode_using_pymatching_v2
        else:
            _method_to_use = decode_using_pymatching_v1

    _method_to_use(
        num_shots=num_shots,
        num_dets=num_dets,
        num_obs=num_obs,
        dem_path=dem_path,
        dets_b8_in_path=dets_b8_in_path,
        obs_predictions_b8_out_path=obs_predictions_b8_out_path,
        tmp_dir=tmp_dir,
    )



import argparse
from typing import Any
from typing import List

from sinter._predict import predict_on_disk


def parse_args(args: List[str]) -> Any:
    parser = argparse.ArgumentParser(
        description='Predict observable flips from detector data.',
        prog='sinter predict',
    )
    parser.add_argument('--dets',
                        type=str,
                        required=True,
                        help='File to read detection event data from.')
    parser.add_argument('--dets_format',
                        type=str,
                        required=True,
                        help='Format detection event data is stored in.\n'
                             'For example: b8 or 01')
    parser.add_argument('--dem',
                        type=str,
                        required=True,
                        help='File to read detector error model from.')
    parser.add_argument('--decoder',
                        type=str,
                        required=True,
                        help='Decoder to use.')
    parser.add_argument('--obs_out',
                        type=str,
                        required=True,
                        help='Location to write predictions from decoder.')
    parser.add_argument('--obs_out_format',
                        type=str,
                        required=True,
                        help='Format to write predictions in.')

    parser.add_argument('--postselect_detectors_with_non_zero_4th_coord',
                        help='Turns on postselection. '
                             'If any detector with a non-zero 4th coordinate fires, the shot is discarded.',
                        action='store_true')
    parser.add_argument('--discards_out',
                        type=str,
                        default=None,
                        help='Location to write whether each shot should be discarded.'
                             'Specified if and only if --postselect_detectors_with_non_zero_4th_coord.')
    parser.add_argument('--discards_out_format',
                        type=str,
                        default=None,
                        help='Format to write discard data in.'
                             'Specified if and only if --postselect_detectors_with_non_zero_4th_coord.')
    result = parser.parse_args(args)

    if result.postselect_detectors_with_non_zero_4th_coord and result.discards_out is None:
        raise ValueError("Must specify --discards_out to record results of --postselect_detectors_with_non_zero_4th_coord.")
    if result.discards_out is not None and result.discards_out_format is None:
        raise ValueError("Must specify --discards_out_format to specify how to record results of --discards_out.")

    return result


def main_predict(*, command_line_args: List[str]):
    parsed = parse_args(command_line_args)
    predict_on_disk(
        decoder=parsed.decoder,
        dem_path=parsed.dem,
        dets_path=parsed.dets,
        dets_format=parsed.dets_format,
        obs_out_path=parsed.obs_out,
        obs_out_format=parsed.obs_out_format,
        postselect_detectors_with_non_zero_4th_coord=parsed.postselect_detectors_with_non_zero_4th_coord,
        discards_out_path=parsed.discards_out,
        discards_out_format=parsed.discards_out_format,
    )



from typing import Callable, TypeVar, List, Any, Iterable, Optional, TYPE_CHECKING, Dict, Union

from sinter._probability_util import fit_binomial, shot_error_rate_to_piece_error_rate

if TYPE_CHECKING:
    import sinter
    import matplotlib.pyplot as plt


MARKERS: str = "ov*sp^<>8PhH+xXDd|" * 100
T = TypeVar('T')
TVal = TypeVar('TVal')
TKey = TypeVar('TKey')


def split_by(vs: Iterable[T], key_func: Callable[[T], Any]) -> List[List[T]]:
    cur_key: Any = None
    out: List[List[T]] = []
    buf: List[T] = []
    for item in vs:
        key = key_func(item)
        if key != cur_key:
            cur_key = key
            if buf:
                out.append(buf)
                buf = []
        buf.append(item)
    if buf:
        out.append(buf)
    return out


def better_sorted_str_terms(val: Any) -> Any:
    if isinstance(val, tuple):
        return tuple(better_sorted_str_terms(e) for e in val)
    if not isinstance(val, str):
        return val
    terms = split_by(val, lambda c: c in '.0123456789')
    result = []
    for term in terms:
        term = ''.join(term)
        if '.' in term:
            try:
                term = float(term)
            except ValueError:
                try:
                    term = tuple(int(e) for e in term.split('.'))
                except ValueError:
                    pass
        else:
            try:
                term = int(term)
            except ValueError:
                pass
        result.append(term)
    return tuple(result)


def group_by(items: Iterable[TVal],
             *,
             key: Callable[[TVal], TKey],
             ) -> Dict[TKey, List[TVal]]:
    """Groups items based on whether they produce the same key from a function.

    Args:
        items: The items to group.
        key: Items that produce the same value from this function get grouped together.

    Returns:
        A dictionary mapping outputs that were produced by the grouping function to
        the list of items that produced that output.

    Examples:
        >>> import sinter
        >>> sinter.group_by([1, 2, 3], key=lambda i: i == 2)
        {False: [1, 3], True: [2]}

        >>> sinter.group_by(range(10), key=lambda i: i % 3)
        {0: [0, 3, 6, 9], 1: [1, 4, 7], 2: [2, 5, 8]}
    """

    result: Dict[TKey, List[TVal]] = {}

    for item in items:
        curve_id = key(item)
        result.setdefault(curve_id, []).append(item)

    return result


TCurveId = TypeVar('TCurveId')


def plot_discard_rate(
        *,
        ax: 'plt.Axes',
        stats: 'Iterable[sinter.TaskStats]',
        x_func: Callable[['sinter.TaskStats'], Any],
        failure_units_per_shot_func: Callable[['sinter.TaskStats'], Any] = lambda _: 1,
        group_func: Callable[['sinter.TaskStats'], TCurveId] = lambda _: None,
        filter_func: Callable[['sinter.TaskStats'], Any] = lambda _: True,
        plot_args_func: Callable[[int, TCurveId], Dict[str, Any]] = lambda _: {},
        highlight_max_likelihood_factor: Optional[float] = 1e3,
) -> None:
    """Plots discard rates in curves with uncertainty highlights.

    Args:
        ax: The plt.Axes to plot onto. For example, the `ax` value from `fig, ax = plt.subplots(1, 1)`.
        stats: The collected statistics to plot.
        x_func: The X coordinate to use for each stat's data point. For example, this could be
            `x_func=lambda stat: stat.json_metadata['physical_error_rate']`.
        failure_units_per_shot_func: How many discard chances there are per shot. This rescales what the
            discard rate means. By default, it is the discard rate per shot, but this allows
            you to instead make it the discard rate per round. For example, if the metadata
            associated with a shot has a field 'r' which is the number of rounds, then this can be
            achieved with `failure_units_per_shot_func=lambda stats: stats.metadata['r']`.
        group_func: Optional. When specified, multiple curves will be plotted instead of one curve.
            The statistics are grouped into curves based on whether or not they get the same result
            out of this function. For example, this could be `group_func=lambda stat: stat.decoder`.
        filter_func: Optional. When specified, some curves will not be plotted.
            The statistics are filtered and only plotted if filter_func(stat) returns True.
            For example, `filter_func=lambda s: s.json_metadata['basis'] == 'x'` would plot only stats
            where the saved metadata indicates the basis was 'x'.
        plot_args_func: Optional. Specifies additional arguments to give the the underlying calls to
            `plot` and `fill_between` used to do the actual plotting. For example, this can be used
            to specify markers and colors. Takes the index of the curve in sorted order and also a
            curve_id (these will be 0 and None respectively if group_func is not specified). For example,
            this could be:

                plot_args_func=lambda index, curve_id: {'color': 'red'
                                                        if curve_id == 'pymatching'
                                                        else 'blue'}

        highlight_max_likelihood_factor: Controls how wide the uncertainty highlight region around curves is.
            Must be 1 or larger. Hypothesis probabilities at most that many times as unlikely as the max likelihood
            hypothesis will be highlighted.
    """
    filtered_stats: List['sinter.TaskStats'] = [
        stat
        for stat in stats
        if filter_func(stat)
    ]

    if not (highlight_max_likelihood_factor >= 1):
        raise ValueError(f"not (highlight_max_likelihood_factor={highlight_max_likelihood_factor} >= 1)")

    curve_groups = group_by(filtered_stats, key=group_func)
    for k, curve_id in enumerate(sorted(curve_groups.keys(), key=better_sorted_str_terms)):
        this_group_stats = sorted(curve_groups[curve_id], key=x_func)

        xs = []
        ys = []
        xs_range = []
        ys_low = []
        ys_high = []
        for stat in this_group_stats:
            x = float(x_func(stat))
            if stat.shots:
                fit = fit_binomial(
                    num_shots=stat.shots,
                    num_hits=stat.discards,
                    max_likelihood_factor=highlight_max_likelihood_factor)
                pieces = failure_units_per_shot_func(stat)
                if stat.discards:
                    xs.append(x)
                    ys.append(shot_error_rate_to_piece_error_rate(fit.best, pieces=pieces))
                xs_range.append(x)
                ys_low.append(shot_error_rate_to_piece_error_rate(fit.low, pieces=pieces))
                ys_high.append(shot_error_rate_to_piece_error_rate(fit.high, pieces=pieces))

        kwargs = dict(plot_args_func(k, curve_id))
        if 'label' not in kwargs and curve_id is not None:
            kwargs['label'] = str(curve_id)
        ax.plot(xs, ys, **kwargs)
        if highlight_max_likelihood_factor > 1:
            if 'zorder' not in kwargs:
                kwargs['zorder'] = 0
            if 'alpha' not in kwargs:
                kwargs['alpha'] = 1
            kwargs['zorder'] -= 100
            kwargs['alpha'] *= 0.25
            if 'marker' in kwargs:
                del kwargs['marker']
            if 'linestyle' in kwargs:
                del kwargs['linestyle']
            if 'label' in kwargs:
                del kwargs['label']
            ax.fill_between(xs_range, ys_low, ys_high, **kwargs)


def plot_error_rate(
        *,
        ax: 'plt.Axes',
        stats: 'Iterable[sinter.TaskStats]',
        x_func: Callable[['sinter.TaskStats'], Any],
        failure_units_per_shot_func: Callable[['sinter.TaskStats'], Any] = lambda _: 1,
        group_func: Callable[['sinter.TaskStats'], TCurveId] = lambda _: None,
        filter_func: Callable[['sinter.TaskStats'], Any] = lambda _: True,
        plot_args_func: Callable[[int, TCurveId], Dict[str, Any]] = lambda _k, _c: {'marker': MARKERS[_k]},
        highlight_max_likelihood_factor: Optional[float] = 1e3,
) -> None:
    """Plots error rates in curves with uncertainty highlights.

    Args:
        ax: The plt.Axes to plot onto. For example, the `ax` value from `fig, ax = plt.subplots(1, 1)`.
        stats: The collected statistics to plot.
        x_func: The X coordinate to use for each stat's data point. For example, this could be
            `x_func=lambda stat: stat.json_metadata['physical_error_rate']`.
        failure_units_per_shot_func: How many error chances there are per shot. This rescales what the
            logical error rate means. By default, it is the logical error rate per shot, but this allows
            you to instead make it the logical error rate per round. For example, if the metadata
            associated with a shot has a field 'r' which is the number of rounds, then this can be
            achieved with `failure_units_per_shot_func=lambda stats: stats.metadata['r']`.
        group_func: Optional. When specified, multiple curves will be plotted instead of one curve.
            The statistics are grouped into curves based on whether or not they get the same result
            out of this function. For example, this could be `group_func=lambda stat: stat.decoder`.
        filter_func: Optional. When specified, some curves will not be plotted.
            The statistics are filtered and only plotted if filter_func(stat) returns True.
            For example, `filter_func=lambda s: s.json_metadata['basis'] == 'x'` would plot only stats
            where the saved metadata indicates the basis was 'x'.
        plot_args_func: Optional. Specifies additional arguments to give the the underlying calls to
            `plot` and `fill_between` used to do the actual plotting. For example, this can be used
            to specify markers and colors. Takes the index of the curve in sorted order and also a
            curve_id (these will be 0 and None respectively if group_func is not specified). For example,
            this could be:

                plot_args_func=lambda index, curve_id: {'color': 'red'
                                                        if curve_id == 'pymatching'
                                                        else 'blue'}

        highlight_max_likelihood_factor: Controls how wide the uncertainty highlight region around curves is.
            Must be 1 or larger. Hypothesis probabilities at most that many times as unlikely as the max likelihood
            hypothesis will be highlighted.
    """
    if not (highlight_max_likelihood_factor >= 1):
        raise ValueError(f"not (highlight_max_likelihood_factor={highlight_max_likelihood_factor} >= 1)")

    filtered_stats: List['sinter.TaskStats'] = [
        stat
        for stat in stats
        if filter_func(stat)
    ]

    curve_groups = group_by(filtered_stats, key=group_func)
    for k, curve_id in enumerate(sorted(curve_groups.keys(), key=better_sorted_str_terms)):
        this_group_stats = sorted(curve_groups[curve_id], key=x_func)

        xs = []
        ys = []
        xs_range = []
        ys_low = []
        ys_high = []
        for stat in this_group_stats:
            num_kept = stat.shots - stat.discards
            if num_kept == 0:
                continue
            x = float(x_func(stat))
            fit = fit_binomial(
                num_shots=num_kept,
                num_hits=stat.errors,
                max_likelihood_factor=highlight_max_likelihood_factor,
            )
            pieces = failure_units_per_shot_func(stat)
            if stat.errors:
                xs.append(x)
                ys.append(shot_error_rate_to_piece_error_rate(fit.best, pieces=pieces))
            if highlight_max_likelihood_factor > 1:
                xs_range.append(x)
                ys_low.append(shot_error_rate_to_piece_error_rate(fit.low, pieces=pieces))
                ys_high.append(shot_error_rate_to_piece_error_rate(fit.high, pieces=pieces))

        kwargs = dict(plot_args_func(k, curve_id))
        if 'label' not in kwargs and curve_id is not None:
            kwargs['label'] = str(curve_id)
        ax.plot(xs, ys, **kwargs)
        if highlight_max_likelihood_factor > 1:
            if 'zorder' not in kwargs:
                kwargs['zorder'] = 0
            if 'alpha' not in kwargs:
                kwargs['alpha'] = 1
            kwargs['zorder'] -= 100
            kwargs['alpha'] *= 0.25
            if 'marker' in kwargs:
                del kwargs['marker']
            if 'linestyle' in kwargs:
                del kwargs['linestyle']
            if 'label' in kwargs:
                del kwargs['label']
            ax.fill_between(xs_range, ys_low, ys_high, **kwargs)



import multiprocessing
import pathlib
import tempfile

import stim

import sinter
from sinter._worker import WorkIn, WorkOut, worker_loop


def test_worker_loop_infers_dem():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        circuit =  stim.Circuit("""
            M(0.2) 0 1
            DETECTOR rec[-1]
            OBSERVABLE_INCLUDE(0) rec[-1] rec[-2]
        """)
        circuit_path = str(tmp_dir / 'input_circuit.stim')
        dem_path = str(tmp_dir / 'input_dem.dem')
        circuit.to_file(circuit_path)
        inp = multiprocessing.Queue()
        out = multiprocessing.Queue()
        inp.put(WorkIn(
            work_key='test1',
            circuit_path=circuit_path,
            dem_path=dem_path,
            decoder='pymatching',
            json_metadata=5,
            strong_id=None,
            num_shots=-1,
            postselected_observables_mask=None,
            postselection_mask=None,
        ))
        inp.put(None)
        worker_loop(tmp_dir, inp, out, 0)
        result: WorkOut = out.get(timeout=1)
        assert out.empty()

        assert result.stats is None
        assert result.work_key == 'test1'
        assert result.msg_error is None
        assert stim.DetectorErrorModel.from_file(dem_path) == circuit.detector_error_model()
        assert result.strong_id is not None


def test_worker_loop_does_not_recompute_dem():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        circuit_path = str(tmp_dir / 'input_circuit.stim')
        dem_path = str(tmp_dir / 'input_dem.dem')
        stim.Circuit("""
            M(0.2) 0 1
            DETECTOR rec[-1]
            OBSERVABLE_INCLUDE(0) rec[-1] rec[-2]
        """).to_file(circuit_path)
        stim.DetectorErrorModel("""
            error(0.234567) D0 L0
        """).to_file(dem_path)

        inp = multiprocessing.Queue()
        out = multiprocessing.Queue()
        inp.put(WorkIn(
            work_key='test1',
            circuit_path=circuit_path,
            dem_path=dem_path,
            decoder='pymatching',
            json_metadata=5,
            strong_id="fake",
            num_shots=1000,
            postselected_observables_mask=None,
            postselection_mask=None,
        ))
        inp.put(None)
        worker_loop(tmp_dir, inp, out, 0)
        result: WorkOut = out.get(timeout=1)
        assert out.empty()

        assert result.stats.shots == 1000
        assert result.stats.discards == 0
        assert 0 < result.stats.errors < 1000
        assert result.work_key == 'test1'
        assert result.msg_error is None
        assert result.strong_id == 'fake'



import os

import contextlib
import functools
import pathlib
import subprocess
import tempfile

import math
import time
from typing import Optional, Dict, Callable, Tuple, Iterable
from typing import Union

import numpy as np
import stim

from sinter._anon_task_stats import AnonTaskStats
from sinter._decoding_fusion_blossom import decode_using_fusion_blossom
from sinter._decoding_internal import decode_using_internal_decoder
from sinter._decoding_pymatching import decode_using_pymatching


DECODER_METHODS: Dict[str, Callable] = {
    'pymatching': decode_using_pymatching,
    'fusion_blossom': decode_using_fusion_blossom,
    'internal': functools.partial(decode_using_internal_decoder,
                                  use_correlated_decoding=False),
    'internal_correlated': functools.partial(decode_using_internal_decoder,
                                             use_correlated_decoding=True),
}


def streaming_post_select(*,
                          num_dets: int,
                          num_obs: int,
                          dets_in_b8: pathlib.Path,
                          obs_in_b8: Optional[pathlib.Path],
                          dets_out_b8: pathlib.Path,
                          obs_out_b8: Optional[pathlib.Path],
                          discards_out_b8: Optional[pathlib.Path],
                          num_shots: int,
                          post_mask: np.ndarray) -> int:
    if post_mask.shape != ((num_dets + 7) // 8,):
        raise ValueError(f"post_mask.shape={post_mask.shape} != (math.ceil(num_detectors / 8),)")
    if post_mask.dtype != np.uint8:
        raise ValueError(f"post_mask.dtype={post_mask.dtype} != np.uint8")
    assert (obs_in_b8 is None) == (obs_out_b8 is None)

    num_det_bytes = math.ceil(num_dets / 8)
    num_obs_bytes = math.ceil(num_obs / 8)
    num_shots_left = num_shots
    num_discards = 0

    with contextlib.ExitStack() as ctx:
        dets_in_f = ctx.enter_context(open(dets_in_b8, 'rb'))
        dets_out_f = ctx.enter_context(open(dets_out_b8, 'wb'))
        if obs_in_b8 is not None and obs_out_b8 is not None:
            obs_in_f = ctx.enter_context(open(obs_in_b8, 'rb'))
            obs_out_f = ctx.enter_context(open(obs_out_b8, 'wb'))
        else:
            obs_in_f = None
            obs_out_f = None
        if discards_out_b8 is not None:
            discards_out_f = ctx.enter_context(open(discards_out_b8, 'wb'))
        else:
            discards_out_f = None

        while num_shots_left:
            batch_size = min(num_shots_left, math.ceil(10 ** 6 / max(1, num_dets)))

            det_batch = np.fromfile(dets_in_f, dtype=np.uint8, count=num_det_bytes * batch_size)
            det_batch.shape = (batch_size, num_det_bytes)
            discarded = np.any(det_batch & post_mask, axis=1)
            det_left = det_batch[~discarded, :]
            det_left.tofile(dets_out_f)

            if obs_in_f is not None and obs_out_f is not None:
                obs_batch = np.fromfile(obs_in_f, dtype=np.uint8, count=num_obs_bytes * batch_size)
                obs_batch.shape = (batch_size, num_obs_bytes)
                obs_left = obs_batch[~discarded, :]
                obs_left.tofile(obs_out_f)
            if discards_out_f is not None:
                discarded.tofile(discards_out_f)

            num_discards += np.count_nonzero(discarded)
            num_shots_left -= batch_size

    return num_discards


def _streaming_count_mistakes(
        *,
        num_shots: int,
        num_obs: int,
        postselected_observable_mask: Optional[np.ndarray] = None,
        obs_in: pathlib.Path,
        predictions_in: pathlib.Path) -> Tuple[int, int]:

    num_obs_bytes = math.ceil(num_obs / 8)
    num_shots_left = num_shots
    num_errors = 0
    num_discards = 0
    with open(obs_in, 'rb') as obs_in_f:
        with open(predictions_in, 'rb') as predictions_in_f:
            while num_shots_left:
                batch_size = min(num_shots_left, math.ceil(10**6 / max(num_obs, 1)))

                obs_batch = np.fromfile(obs_in_f, dtype=np.uint8, count=num_obs_bytes * batch_size)
                pred_batch = np.fromfile(predictions_in_f, dtype=np.uint8, count=num_obs_bytes * batch_size)
                obs_batch.shape = (batch_size, num_obs_bytes)
                pred_batch.shape = (batch_size, num_obs_bytes)

                cmp_table = pred_batch ^ obs_batch
                if postselected_observable_mask is None:
                    local_discards = 0
                else:
                    local_discards = np.count_nonzero(np.any(cmp_table & postselected_observable_mask, axis=1))
                local_errors_or_discards = np.count_nonzero(np.any(cmp_table, axis=1))
                num_discards += local_discards
                num_errors += local_errors_or_discards - local_discards
                num_shots_left -= batch_size
    return num_discards, num_errors


def sample_decode(*,
                  circuit_obj: Optional[stim.Circuit],
                  circuit_path: Union[None, str, pathlib.Path],
                  dem_obj: Optional[stim.DetectorErrorModel],
                  dem_path: Union[None, str, pathlib.Path],
                  post_mask: Optional[np.ndarray] = None,
                  postselected_observable_mask: Optional[np.ndarray] = None,
                  num_shots: int,
                  decoder: str,
                  tmp_dir: Union[str, pathlib.Path, None] = None) -> AnonTaskStats:
    """Samples how many times a decoder correctly predicts the logical frame.

    Args:
        circuit_obj: The noisy circuit to sample from and decode results for.
            Must specify circuit_obj XOR circuit_path.
        circuit_path: The file storing the circuit to sample from.
            Must specify circuit_obj XOR circuit_path.
        dem_obj: The error model to give to the decoder.
            Must specify dem_obj XOR dem_path.
        dem_path: The file storing the error model to give to the decoder.
            Must specify dem_obj XOR dem_path.
        post_mask: Postselection mask. Any samples that have a non-zero result
            at a location where the mask has a 1 bit are discarded. If set to
            None, no postselection is performed.
        postselected_observable_mask: Bit packed mask indicating which observables to
            postselect on. If the decoder incorrectly predicts any of these observables, the
            shot is discarded instead of counted as an error.
        num_shots: The number of sample shots to take from the circuit.
        decoder: The name of the decoder to use. Allowed values are:
            "pymatching":
                Use pymatching min-weight-perfect-match decoder.
            "internal":
                Use internal decoder with uncorrelated decoding.
            "internal_correlated":
                Use internal decoder with correlated decoding.
        tmp_dir: An existing directory that is currently empty where temporary
            files can be written as part of performing decoding. If set to
            None, one is created using the tempfile package.
    """
    if (circuit_obj is None) == (circuit_path is None):
        raise ValueError('(circuit_obj is None) == (circuit_path is None)')
    if (dem_obj is None) == (dem_path is None):
        raise ValueError('(dem_obj is None) == (dem_path is None)')
    if num_shots == 0:
        return AnonTaskStats()

    with contextlib.ExitStack() as exit_stack:
        start_time = time.monotonic()

        if circuit_path is not None:
            circuit = stim.Circuit.from_file(circuit_path)
        else:
            circuit = circuit_obj
        if tmp_dir is None:
            tmp_dir = exit_stack.enter_context(tempfile.TemporaryDirectory())
        tmp_dir = pathlib.Path(tmp_dir)
        if dem_path is None:
            dem_path = tmp_dir / 'tmp.dem'
            dem_obj.to_file(dem_path)
        dem_path = pathlib.Path(dem_path)

        dets_all_path = tmp_dir / 'sinter_dets.all.b8'
        obs_all_path = tmp_dir / 'sinter_obs.all.b8'
        dets_kept_path = tmp_dir / 'sinter_dets.kept.b8'
        obs_kept_path = tmp_dir / 'sinter_obs.kept.b8'
        predictions_path = tmp_dir / 'sinter_predictions.b8'

        num_dets = circuit.num_detectors
        num_obs = circuit.num_observables

        # Sample data using Stim.
        sampler: stim.CompiledDetectorSampler = circuit.compile_detector_sampler()
        sampler.sample_write(
            num_shots,
            filepath=str(dets_all_path),
            obs_out_filepath=str(obs_all_path),
            format='b8',
            obs_out_format='b8',
        )

        # Postselect, then split into detection event data and observable data.
        if post_mask is None:
            num_det_discards = 0
            dets_used_path = dets_all_path
            obs_used_path = obs_all_path
        else:
            num_det_discards = streaming_post_select(
                num_shots=num_shots,
                num_dets=num_dets,
                num_obs=num_obs,
                dets_in_b8=dets_all_path,
                dets_out_b8=dets_kept_path,
                obs_in_b8=obs_all_path,
                obs_out_b8=obs_kept_path,
                post_mask=post_mask,
                discards_out_b8=None,
            )
            dets_used_path = dets_kept_path
            obs_used_path = obs_kept_path
        num_kept_shots = num_shots - num_det_discards

        # Perform syndrome decoding to predict observables from detection events.
        decode_method = DECODER_METHODS.get(decoder)
        if decode_method is None:
            raise NotImplementedError(f"Unrecognized decoder: {decoder!r}")
        decode_method(
            num_shots=num_kept_shots,
            num_dets=num_dets,
            num_obs=num_obs,
            dem_path=dem_path,
            dets_b8_in_path=dets_used_path,
            obs_predictions_b8_out_path=predictions_path,
            tmp_dir=tmp_dir,
        )

        # Count how many predictions matched the actual observable data.
        num_obs_discards, num_errors = _streaming_count_mistakes(
            num_shots=num_kept_shots,
            num_obs=num_obs,
            obs_in=obs_used_path,
            predictions_in=predictions_path,
            postselected_observable_mask=postselected_observable_mask,
        )

        return AnonTaskStats(
            shots=num_shots,
            errors=num_errors,
            discards=num_obs_discards + num_det_discards,
            seconds=time.monotonic() - start_time,
        )
